def test_landing_page(app_client):
    """Test landing page"""
    resp = app_client.get("/")
    assert resp.status_code == 200
    resp_json = resp.json()

    # Make sure OpenAPI docs are linked
    docs = next(filter(lambda link: link["rel"] == "docs", resp_json["links"]))["href"]
    resp = app_client.get(docs)
    assert resp.status_code == 200

    # Make sure conformance classes are linked
    conf = next(filter(lambda link: link["rel"] == "conformance", resp_json["links"]))[
        "href"
    ]
    resp = app_client.get(conf)
    assert resp.status_code == 200



"""Application settings."""
import enum
from typing import List, Optional, Set

from pydantic import BaseSettings


# TODO: Move to stac-pydantic
class ApiExtensions(enum.Enum):
    """
    Enumeration of available stac api extensions
    Ref: https://github.com/radiantearth/stac-api-spec/tree/master/extensions
    """

    context = "context"
    fields = "fields"
    query = "query"
    sort = "sort"
    transaction = "transaction"


class AddOns(enum.Enum):
    """
    Enumeration of available third party add ons
    """

    tiles = "tiles"


class ApiSettings(BaseSettings):
    """Application settings"""

    environment: str
    debug: bool = False
    postgres_user: str
    postgres_pass: str
    postgres_host_reader: str
    postgres_host_writer: str
    postgres_port: str
    postgres_dbname: str

    # Enabled api extensions
    stac_api_extensions: Optional[List[ApiExtensions]] = None

    # Enabled third party add ons
    add_ons: Optional[List[AddOns]] = None

    # Fields which are defined by STAC but not included in the database model
    forbidden_fields: Set[str] = {"type"}

    # Fields which are item properties but indexed as distinct fields in the database model
    indexed_fields: Set[str] = {"datetime"}

    # Fields which are always included in the response (fields extension)
    default_includes: Set[str] = {
        "id",
        "type",
        "geometry",
        "bbox",
        "links",
        "assets",
        "properties.datetime",
    }

    class Config:
        """model config"""

        env_file = ".env"

    @property
    def reader_connection_string(self):
        """Create reader psql connection string"""
        return f"postgresql://{self.postgres_user}:{self.postgres_pass}@{self.postgres_host_reader}:{self.postgres_port}/{self.postgres_dbname}"

    @property
    def writer_connection_string(self):
        """Create writer psql connection string"""
        return f"postgresql://{self.postgres_user}:{self.postgres_pass}@{self.postgres_host_writer}:{self.postgres_port}/{self.postgres_dbname}"

    def api_extension_is_enabled(self, ext: ApiExtensions) -> bool:
        """Helper method to check if an api extension is enabled"""
        return (
            ext in self.stac_api_extensions or False
            if self.stac_api_extensions
            else False
        )

    def add_on_is_enabled(self, add_on: AddOns) -> bool:
        """Helper method to check if an add on is enabled"""
        return add_on in self.add_ons or False if self.add_ons else False


settings: Optional[ApiSettings] = None


def inject_settings(base_settings: ApiSettings):
    """Inject settings to global scope"""
    global settings
    settings = base_settings



"""FastAPI application."""
from stac_api.api import create_app
from stac_api.config import ApiSettings

settings = ApiSettings(
    stac_api_extensions=["context", "fields", "query", "sort", "transaction"],
    add_ons=["tiles"],
)
app = create_app(settings=settings)



"""Model serialization."""
import json
from typing import Any, Dict, List, Union
from urllib.parse import urljoin

import geoalchemy2 as ga
from pydantic import BaseModel
from pydantic.utils import GetterDict
from stac_api import config
from stac_api.errors import DatabaseError
from stac_api.models.links import CollectionLinks, ItemLinks, filter_links
from stac_pydantic.shared import DATETIME_RFC339


def resolve_links(links: list, base_url: str) -> List[Dict]:
    """
    Convert relative links to absolute links using the specified base url.  It would be more appropriate to use a view,
    but SQLAlchemy ORM doesn't support this as far as I know.
    """
    if isinstance(links[0], BaseModel):
        links = [link.dict() for link in links]
    filtered_links = filter_links(links)
    for link in filtered_links:
        link.update({"href": urljoin(base_url, link["href"])})
    return filtered_links


class ItemGetter(GetterDict):
    """
    Custom GetterDict used internally by pydantic ORM mode when decomposing database model to pydantic model.  This
    object resolves structural differences between the two models, for example:
      - relative links stored in the database must be resolved absolute links and inferred links must be added
      - ``datetime`` is defined as its own field in the database but as ``item.properties.datetime`` in the stac spec
      - ``geometry`` can be one of several formats when exported from the database but the STAC item expects geojson
    """

    @staticmethod
    def decode_geom(geom: Union[ga.elements.WKBElement, str, Dict]) -> Dict:
        """Decode geoalchemy type to geojson"""
        if isinstance(geom, ga.elements.WKBElement):
            return json.loads(json.dumps(ga.shape.to_shape(geom).__geo_interface__))
        elif isinstance(geom, str):
            return json.loads(geom)
        elif isinstance(geom, dict):
            return geom
        raise DatabaseError("Received unexpected geometry format from database")

    def __init__(self, obj: Any):
        """Decompose orm model to pydantic model"""
        properties = obj.properties.copy()
        for field in config.settings.indexed_fields:
            # Use getattr to accommodate extension namespaces
            field_value = getattr(obj, field.split(":")[-1])
            if field == "datetime":
                field_value = field_value.strftime(DATETIME_RFC339)
            properties[field] = field_value
        # Create inferred links
        item_links = ItemLinks(
            collection_id=obj.collection_id, base_url=obj.base_url, item_id=obj.id
        ).create_links()
        # Resolve existing links
        if obj.links:
            item_links += resolve_links(obj.links, obj.base_url)
        db_model = obj.__class__(
            id=obj.id,
            stac_version=obj.stac_version,
            geometry=self.decode_geom(obj.geometry),
            bbox=obj.bbox,
            properties=properties,
            assets=obj.assets,
            collection_id=obj.collection_id,
            datetime=obj.datetime,
            links=item_links,
            stac_extensions=obj.stac_extensions,
        )
        db_model.type = "Feature"
        db_model.collection = db_model.collection_id
        super().__init__(db_model)


class CollectionGetter(GetterDict):
    """
    Custom GetterDict used internally by pydantic ORM mode when collection ORM model to pydantic model
    """

    def __init__(self, obj: Any):
        """Decompose orm model to pydantic model"""
        # Create inferred links
        collection_links = CollectionLinks(
            collection_id=obj.id, base_url=obj.base_url
        ).create_links()
        # Resolve existing links
        if obj.links:
            collection_links += resolve_links(obj.links, obj.base_url)
        # TODO: Fix bug in stac-pydantic (collection root validator) that requires coercing to empty list
        stac_extensions = obj.stac_extensions or []
        db_model = obj.__class__(
            id=obj.id,
            stac_extensions=stac_extensions,
            stac_version=obj.stac_version,
            title=obj.title,
            description=obj.description,
            keywords=obj.keywords,
            license=obj.license,
            providers=obj.providers,
            summaries=obj.summaries,
            extent=obj.extent,
            links=collection_links,
        )
        super().__init__(db_model)



"""api request/response models"""

import abc
from dataclasses import dataclass
from typing import Dict, Optional, Type, Union

from fastapi import Body, Path

from pydantic import BaseModel, create_model
from pydantic.fields import UndefinedType
from stac_api.config import ApiExtensions, ApiSettings


def _create_request_model(
    model: Type[BaseModel], settings: ApiSettings
) -> Type[BaseModel]:
    """Create a pydantic model for validating a request body"""

    fields = {}
    for (k, v) in model.__fields__.items():
        if k == "query":
            if not settings.api_extension_is_enabled(ApiExtensions.query):
                continue

        if k == "sortby":
            if not settings.api_extension_is_enabled(ApiExtensions.sort):
                continue

        if k == "field":
            if not settings.api_extension_is_enabled(ApiExtensions.fields):
                continue

        field_info = v.field_info
        body = Body(
            None
            if isinstance(field_info.default, UndefinedType)
            else field_info.default,
            default_factory=field_info.default_factory,
            alias=field_info.alias,
            alias_priority=field_info.alias_priority,
            title=field_info.title,
            description=field_info.description,
            const=field_info.const,
            gt=field_info.gt,
            ge=field_info.ge,
            lt=field_info.lt,
            le=field_info.le,
            multiple_of=field_info.multiple_of,
            min_items=field_info.min_items,
            max_items=field_info.max_items,
            min_length=field_info.min_length,
            max_length=field_info.max_length,
            regex=field_info.regex,
            extra=field_info.extra,
        )
        fields[k] = (v.outer_type_, body)
    return create_model(model.__name__, **fields, __base__=model)


@dataclass  # type:ignore
class APIRequest(abc.ABC):
    """Generic API Request base class"""

    @abc.abstractmethod
    def kwargs(self) -> Dict:
        """Transform api request params into format which matches the signature of the endpoint"""
        ...


@dataclass  # type:ignore
class CollectionUri(APIRequest):
    """Delete collection"""

    collectionId: str = Path(..., description="Collection ID")

    def kwargs(self) -> Dict:
        """kwargs"""
        return {"id": self.collectionId}


@dataclass
class ItemUri(CollectionUri):
    """Delete item"""

    itemId: str = Path(..., description="Item ID")

    def kwargs(self) -> Dict:
        """kwargs"""
        return {"id": self.itemId}


@dataclass
class EmptyRequest(APIRequest):
    """Empty request"""

    def kwargs(self) -> Dict:
        """kwargs"""
        return {}


@dataclass
class ItemCollectionUri(CollectionUri):
    """Get item collection"""

    limit: int = 10
    token: str = None

    def kwargs(self) -> Dict:
        """kwargs"""
        return {"id": self.collectionId, "limit": self.limit, "token": self.token}


@dataclass
class SearchGetRequest(APIRequest):
    """GET search request"""

    collections: Optional[str] = None
    ids: Optional[str] = None
    bbox: Optional[str] = None
    datetime: Optional[Union[str]] = None
    limit: Optional[int] = 10
    query: Optional[str] = None
    token: Optional[str] = None
    fields: Optional[str] = None
    sortby: Optional[str] = None

    def kwargs(self) -> Dict:
        """kwargs"""
        return {
            "collections": self.collections.split(",")
            if self.collections
            else self.collections,
            "ids": self.ids.split(",") if self.ids else self.ids,
            "bbox": self.bbox.split(",") if self.bbox else self.bbox,
            "datetime": self.datetime,
            "limit": self.limit,
            "query": self.query,
            "token": self.token,
            "fields": self.fields.split(",") if self.fields else self.fields,
            "sortby": self.sortby.split(",") if self.sortby else self.sortby,
        }



"""Postgresql base client"""
import abc
import logging
from dataclasses import dataclass
from typing import Optional, Type

import sqlalchemy as sa
from sqlalchemy.orm import Query

import psycopg2
from stac_api import errors
from stac_api.models import database
from stac_api.utils.dependencies import READER, WRITER

logger = logging.getLogger(__name__)


@dataclass
class PostgresClient(abc.ABC):
    """Database CRUD operations on the defined table"""

    table: Type[database.BaseModel]

    @property
    def reader_session(self):
        """Get reader session from context var"""
        return READER.get()

    @property
    def writer_session(self):
        """Get writer session from context var"""
        return WRITER.get()

    @staticmethod
    def row_exists(query: Query) -> bool:
        """Check if a record exists from the sqlalchemy query object"""
        return True if query.scalar() else False

    def commit(self) -> None:
        """Commit both reader and writer sessions to keep them in sync, rolling back on psycopg2 errors"""
        try:
            self.writer_session.commit()
            self.reader_session.commit()
        except sa.exc.IntegrityError as e:
            self.writer_session.rollback()
            self.reader_session.rollback()
            logger.error(e.orig.pgerror, exc_info=True)
            # Explicitly catch foreign key errors to be reraised by the API as validation errors
            if isinstance(e.orig, psycopg2.errors.ForeignKeyViolation):
                raise errors.ForeignKeyError(e.orig.pgerror)
            raise errors.DatabaseError(e.orig.pgerror) from e
        except Exception as e:
            logger.error(e, exc_info=True)
            raise errors.DatabaseError("Unhandled database exception during commit")

    def lookup_id(
        self, item_id: str, table: Optional[Type[database.BaseModel]] = None
    ) -> Query:
        """Create a query to access a single record from the table"""
        table = table or self.table
        try:
            query = self.reader_session.query(table).filter(table.id == item_id)
        except Exception as e:
            logger.error(e, exc_info=True)
            raise errors.DatabaseError("Unhandled database during ID lookup")
        if not self.row_exists(query):
            error_message = f"Row {item_id} does not exist"
            logger.warning(error_message)
            raise errors.NotFoundError(error_message)
        return query



from datetime import datetime, timedelta

from starlette.testclient import TestClient

from stac_api.api import create_app
from stac_api.api.routers import create_core_router, create_transactions_router
from stac_api.models.schemas import Collection, Item

from ..conftest import ApiSettings, MockStarletteRequest

STAC_CORE_ROUTES = [
    "GET /",
    "GET /collections",
    "GET /collections/{collectionId}",
    "GET /collections/{collectionId}/items",
    "GET /collections/{collectionId}/items/{itemId}",
    "GET /conformance",
    "GET /search",
    "POST /search",
]

STAC_TRANSACTION_ROUTES = [
    "DELETE /collections/{collectionId}",
    "DELETE /collections/{collectionId}/items/{itemId}",
    "POST /collections",
    "POST /collections/{collectionId}/items",
    "PUT /collections",
    "PUT /collections/{collectionId}/items",
]


def test_core_router(postgres_core):
    settings = ApiSettings()
    router = create_core_router(postgres_core, settings)
    assert STAC_CORE_ROUTES == sorted(
        [f"{list(route.methods)[0]} {route.path}" for route in router.routes]
    )


def test_transactions_router(postgres_transactions):
    settings = ApiSettings()
    router = create_transactions_router(postgres_transactions, settings)
    assert STAC_TRANSACTION_ROUTES == sorted(
        [f"{list(route.methods)[0]} {route.path}" for route in router.routes]
    )


def test_app_transaction_extension(load_test_data):
    settings = ApiSettings(stac_api_extensions=["transaction"])
    app = create_app(settings)

    with TestClient(app) as client:
        collection = load_test_data("test_collection.json")
        resp = client.post("/collections", json=collection)
        assert resp.status_code == 200

        item = load_test_data("test_item.json")
        resp = client.post(f"/collections/{item['collection']}/items", json=item)
        assert resp.status_code == 200


def test_app_context_extension(load_test_data, postgres_transactions):
    settings = ApiSettings(stac_api_extensions=["context"])
    app = create_app(settings)

    coll = Collection.parse_obj(load_test_data("test_collection.json"))
    item = Item.parse_obj(load_test_data("test_item.json"))
    postgres_transactions.create_collection(coll, request=MockStarletteRequest)
    postgres_transactions.create_item(item, request=MockStarletteRequest)

    with TestClient(app) as client:
        resp = client.get("/search", params={"collections": ["test-collection"]})
        assert resp.status_code == 200
        resp_json = resp.json()

        assert "context" in resp_json
        assert resp_json["context"]["returned"] == resp_json["context"]["matched"] == 1


def test_app_fields_extension(load_test_data, postgres_transactions):
    settings = ApiSettings(stac_api_extensions=["fields"])
    app = create_app(settings)

    coll = Collection.parse_obj(load_test_data("test_collection.json"))
    item = Item.parse_obj(load_test_data("test_item.json"))
    postgres_transactions.create_collection(coll, request=MockStarletteRequest)
    postgres_transactions.create_item(item, request=MockStarletteRequest)

    with TestClient(app) as client:
        resp = client.get("/search", params={"collections": ["test-collection"]})
        assert resp.status_code == 200
        resp_json = resp.json()
        assert list(resp_json["features"][0]["properties"]) == ["datetime"]


def test_app_query_extension(load_test_data, postgres_transactions):
    settings = ApiSettings(stac_api_extensions=["query"],)
    app = create_app(settings)

    coll = Collection.parse_obj(load_test_data("test_collection.json"))
    test_item = load_test_data("test_item.json")
    item = Item.parse_obj(test_item)
    postgres_transactions.create_collection(coll, request=MockStarletteRequest)
    postgres_transactions.create_item(item, request=MockStarletteRequest)

    with TestClient(app) as client:
        params = {
            "query": {"proj:epsg": {"gt": test_item["properties"]["proj:epsg"] + 1}}
        }
        resp = client.post("/search", json=params)
        assert resp.status_code == 200
        resp_json = resp.json()
        assert len(resp_json["features"]) == 0


def test_app_sort_extension(load_test_data, postgres_transactions):
    settings = ApiSettings(stac_api_extensions=["sort"],)
    app = create_app(settings)

    coll = Collection.parse_obj(load_test_data("test_collection.json"))
    postgres_transactions.create_collection(coll, request=MockStarletteRequest)

    first_item = load_test_data("test_item.json")
    item_date = datetime.strptime(
        first_item["properties"]["datetime"], "%Y-%m-%dT%H:%M:%SZ"
    )
    postgres_transactions.create_item(
        Item.parse_obj(first_item), request=MockStarletteRequest
    )

    second_item = load_test_data("test_item.json")
    second_item["id"] = "another-item"
    another_item_date = item_date - timedelta(days=1)
    second_item["properties"]["datetime"] = another_item_date.strftime(
        "%Y-%m-%dT%H:%M:%SZ"
    )
    postgres_transactions.create_item(
        Item.parse_obj(second_item), request=MockStarletteRequest
    )

    with TestClient(app) as client:
        params = {
            "collections": [first_item["collection"]],
            "sortby": [{"field": "datetime", "direction": "desc"}],
        }
        resp = client.post("/search", json=params)
        assert resp.status_code == 200
        resp_json = resp.json()
        assert resp_json["features"][0]["id"] == first_item["id"]
        assert resp_json["features"][1]["id"] == second_item["id"]



import json
import time
import uuid
from copy import deepcopy
from datetime import datetime, timedelta
from random import randint
from urllib.parse import parse_qs, urlparse, urlsplit

from shapely.geometry import Polygon

from stac_pydantic.api.search import DATETIME_RFC339


def test_create_and_delete_item(app_client, load_test_data):
    """Test creation and deletion of a single item (transactions extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    resp = app_client.delete(
        f"/collections/{test_item['collection']}/items/{resp.json()['id']}"
    )
    assert resp.status_code == 200


def test_create_item_conflict(app_client, load_test_data):
    """Test creation of an item which already exists (transactions extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 409


def test_delete_missing_item(app_client, load_test_data):
    """Test deletion of an item which does not exist (transactions extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.delete(f"/collections/{test_item['collection']}/items/hijosh")
    assert resp.status_code == 404


def test_create_item_missing_collection(app_client, load_test_data):
    """Test creation of an item without a parent collection (transactions extension)"""
    test_item = load_test_data("test_item.json")
    test_item["collection"] = "stac is cool"
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 422


def test_update_item_already_exists(app_client, load_test_data):
    """Test updating an item which already exists (transactions extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    assert test_item["properties"]["gsd"] != 16
    test_item["properties"]["gsd"] = 16
    resp = app_client.put(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    updated_item = resp.json()
    assert updated_item["properties"]["gsd"] == 16


def test_update_new_item(app_client, load_test_data):
    """Test updating an item which does not exist (transactions extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.put(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200


def test_update_item_missing_collection(app_client, load_test_data):
    """Test updating an item without a parent collection (transactions extension)"""
    test_item = load_test_data("test_item.json")
    test_item["collection"] = "stac is cool"
    resp = app_client.put(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 422


def test_get_item(app_client, load_test_data):
    """Test read an item by id (core)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    get_item = app_client.get(
        f"/collections/{test_item['collection']}/items/{test_item['id']}"
    )
    assert get_item.status_code == 200


def test_get_item_collection(app_client, load_test_data):
    """Test read an item collection (core)"""
    item_count = randint(1, 4)
    test_item = load_test_data("test_item.json")

    for idx in range(item_count):
        _test_item = deepcopy(test_item)
        _test_item["id"] = test_item["id"] + str(idx)
        resp = app_client.post(
            f"/collections/{test_item['collection']}/items", json=_test_item
        )
        assert resp.status_code == 200

    resp = app_client.get(f"/collections/{test_item['collection']}/items")
    assert resp.status_code == 200

    item_collection = resp.json()
    assert item_collection["context"]["matched"] == len(range(item_count))


def test_pagination(app_client, load_test_data):
    """Test item collection pagination (paging extension)"""
    item_count = 10
    test_item = load_test_data("test_item.json")

    for idx in range(item_count):
        _test_item = deepcopy(test_item)
        _test_item["id"] = test_item["id"] + str(idx)
        resp = app_client.post(
            f"/collections/{test_item['collection']}/items", json=_test_item
        )
        assert resp.status_code == 200

    resp = app_client.get(
        f"/collections/{test_item['collection']}/items", params={"limit": 3}
    )
    assert resp.status_code == 200
    first_page = resp.json()
    assert first_page["context"]["returned"] == 3

    url_components = urlsplit(first_page["links"][0]["href"])
    resp = app_client.get(f"{url_components.path}?{url_components.query}")
    assert resp.status_code == 200
    second_page = resp.json()
    assert second_page["context"]["returned"] == 3


def test_item_timestamps(app_client, load_test_data):
    """Test created and updated timestamps (common metadata)"""
    test_item = load_test_data("test_item.json")
    start_time = datetime.utcnow()
    time.sleep(2)
    # Confirm `created` timestamp
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    item = resp.json()
    created_dt = datetime.strptime(item["properties"]["created"], DATETIME_RFC339)
    assert resp.status_code == 200
    assert start_time < created_dt < datetime.utcnow()

    time.sleep(2)
    # Confirm `updated` timestamp
    item["properties"]["proj:epsg"] = 4326
    resp = app_client.put(f"/collections/{test_item['collection']}/items", json=item)
    assert resp.status_code == 200
    updated_item = resp.json()

    # Created shouldn't change on update
    assert item["properties"]["created"] == updated_item["properties"]["created"]
    assert (
        datetime.strptime(updated_item["properties"]["updated"], DATETIME_RFC339)
        > created_dt
    )


def test_item_search_by_id_post(app_client, load_test_data):
    """Test POST search by item id (core)"""
    ids = ["test1", "test2", "test3"]
    for id in ids:
        test_item = load_test_data("test_item.json")
        test_item["id"] = id
        resp = app_client.post(
            f"/collections/{test_item['collection']}/items", json=test_item
        )
        assert resp.status_code == 200

    params = {"collections": [test_item["collection"]], "ids": ids}
    resp = app_client.post("/search", json=params)
    assert resp.status_code == 200
    resp_json = resp.json()
    assert len(resp_json["features"]) == len(ids)
    assert set([feat["id"] for feat in resp_json["features"]]) == set(ids)


def test_item_search_spatial_query_post(app_client, load_test_data):
    """Test POST search with spatial query (core)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    params = {
        "collections": [test_item["collection"]],
        "intersects": test_item["geometry"],
    }
    resp = app_client.post("/search", json=params)
    assert resp.status_code == 200
    resp_json = resp.json()
    assert resp_json["features"][0]["id"] == test_item["id"]


def test_item_search_temporal_query_post(app_client, load_test_data):
    """Test POST search with single-tailed spatio-temporal query (core)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    item_date = datetime.strptime(test_item["properties"]["datetime"], DATETIME_RFC339)
    item_date = item_date + timedelta(seconds=1)

    params = {
        "collections": [test_item["collection"]],
        "intersects": test_item["geometry"],
        "datetime": item_date.strftime(DATETIME_RFC339),
    }
    resp = app_client.post("/search", json=params)
    resp_json = resp.json()
    assert resp_json["features"][0]["id"] == test_item["id"]


def test_item_search_temporal_window_post(app_client, load_test_data):
    """Test POST search with two-tailed spatio-temporal query (core)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    item_date = datetime.strptime(test_item["properties"]["datetime"], DATETIME_RFC339)
    item_date_before = item_date - timedelta(seconds=1)
    item_date_after = item_date + timedelta(seconds=1)

    params = {
        "collections": [test_item["collection"]],
        "intersects": test_item["geometry"],
        "datetime": f"{item_date_before.strftime(DATETIME_RFC339)}/{item_date_after.strftime(DATETIME_RFC339)}",
    }
    resp = app_client.post("/search", json=params)
    resp_json = resp.json()
    assert resp_json["features"][0]["id"] == test_item["id"]


def test_item_search_temporal_open_window(app_client, load_test_data):
    """Test POST search with open spatio-temporal query (core)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    params = {
        "collections": [test_item["collection"]],
        "intersects": test_item["geometry"],
        "datetime": "../..",
    }
    resp = app_client.post("/search", json=params)
    resp_json = resp.json()
    assert resp_json["features"][0]["id"] == test_item["id"]


def test_item_search_sort_post(app_client, load_test_data):
    """Test POST search with sorting (sort extension)"""
    first_item = load_test_data("test_item.json")
    item_date = datetime.strptime(first_item["properties"]["datetime"], DATETIME_RFC339)
    resp = app_client.post(
        f"/collections/{first_item['collection']}/items", json=first_item
    )
    assert resp.status_code == 200

    second_item = load_test_data("test_item.json")
    second_item["id"] = "another-item"
    another_item_date = item_date - timedelta(days=1)
    second_item["properties"]["datetime"] = another_item_date.strftime(DATETIME_RFC339)
    resp = app_client.post(
        f"/collections/{second_item['collection']}/items", json=second_item
    )
    assert resp.status_code == 200

    params = {
        "collections": [first_item["collection"]],
        "sortby": [{"field": "datetime", "direction": "desc"}],
    }
    resp = app_client.post("/search", json=params)
    assert resp.status_code == 200
    resp_json = resp.json()
    assert resp_json["features"][0]["id"] == first_item["id"]
    assert resp_json["features"][1]["id"] == second_item["id"]


def test_item_search_by_id_get(app_client, load_test_data):
    """Test GET search by item id (core)"""
    ids = ["test1", "test2", "test3"]
    for id in ids:
        test_item = load_test_data("test_item.json")
        test_item["id"] = id
        resp = app_client.post(
            f"/collections/{test_item['collection']}/items", json=test_item
        )
        assert resp.status_code == 200

    params = {"collections": test_item["collection"], "ids": ",".join(ids)}
    resp = app_client.get("/search", params=params)
    assert resp.status_code == 200
    resp_json = resp.json()
    assert len(resp_json["features"]) == len(ids)
    assert set([feat["id"] for feat in resp_json["features"]]) == set(ids)


def test_item_search_bbox_get(app_client, load_test_data):
    """Test GET search with spatial query (core)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    params = {
        "collections": test_item["collection"],
        "bbox": ",".join([str(coord) for coord in test_item["bbox"]]),
    }
    resp = app_client.get("/search", params=params)
    assert resp.status_code == 200
    resp_json = resp.json()
    assert resp_json["features"][0]["id"] == test_item["id"]


def test_item_search_get_without_collections(app_client, load_test_data):
    """Test GET search without specifying collections"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    params = {
        "bbox": ",".join([str(coord) for coord in test_item["bbox"]]),
    }
    resp = app_client.get("/search", params=params)
    assert resp.status_code == 200
    resp_json = resp.json()
    assert resp_json["features"][0]["id"] == test_item["id"]


def test_item_search_temporal_window_get(app_client, load_test_data):
    """Test GET search with spatio-temporal query (core)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    item_date = datetime.strptime(test_item["properties"]["datetime"], DATETIME_RFC339)
    item_date_before = item_date - timedelta(seconds=1)
    item_date_after = item_date + timedelta(seconds=1)

    params = {
        "collections": test_item["collection"],
        "bbox": ",".join([str(coord) for coord in test_item["bbox"]]),
        "datetime": f"{item_date_before.strftime(DATETIME_RFC339)}/{item_date_after.strftime(DATETIME_RFC339)}",
    }
    resp = app_client.get("/search", params=params)
    resp_json = resp.json()
    assert resp_json["features"][0]["id"] == test_item["id"]


def test_item_search_sort_get(app_client, load_test_data):
    """Test GET search with sorting (sort extension)"""
    first_item = load_test_data("test_item.json")
    item_date = datetime.strptime(first_item["properties"]["datetime"], DATETIME_RFC339)
    resp = app_client.post(
        f"/collections/{first_item['collection']}/items", json=first_item
    )
    assert resp.status_code == 200

    second_item = load_test_data("test_item.json")
    second_item["id"] = "another-item"
    another_item_date = item_date - timedelta(days=1)
    second_item["properties"]["datetime"] = another_item_date.strftime(DATETIME_RFC339)
    resp = app_client.post(
        f"/collections/{second_item['collection']}/items", json=second_item
    )
    assert resp.status_code == 200
    params = {"collections": [first_item["collection"]], "sortby": "-datetime"}
    resp = app_client.get("/search", params=params)
    assert resp.status_code == 200
    resp_json = resp.json()
    assert resp_json["features"][0]["id"] == first_item["id"]
    assert resp_json["features"][1]["id"] == second_item["id"]


def test_item_search_post_without_collection(app_client, load_test_data):
    """Test POST search without specifying a collection"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    params = {
        "bbox": test_item["bbox"],
    }
    resp = app_client.post("/search", json=params)
    assert resp.status_code == 200
    resp_json = resp.json()
    assert resp_json["features"][0]["id"] == test_item["id"]


def test_item_search_properties_jsonb(app_client, load_test_data):
    """Test POST search with JSONB query (query extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    # EPSG is a JSONB key
    params = {"query": {"proj:epsg": {"gt": test_item["properties"]["proj:epsg"] + 1}}}
    resp = app_client.post("/search", json=params)
    assert resp.status_code == 200
    resp_json = resp.json()
    assert len(resp_json["features"]) == 0


def test_item_search_properties_field(app_client, load_test_data):
    """Test POST search indexed field with query (query extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    # Orientation is an indexed field
    params = {"query": {"orientation": {"eq": "south"}}}
    resp = app_client.post("/search", json=params)
    assert resp.status_code == 200
    resp_json = resp.json()
    assert len(resp_json["features"]) == 0


def test_item_search_get_query_extension(app_client, load_test_data):
    """Test GET search with JSONB query (query extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    # EPSG is a JSONB key
    params = {
        "collections": [test_item["collection"]],
        "query": json.dumps(
            {"proj:epsg": {"gt": test_item["properties"]["proj:epsg"] + 1}}
        ),
    }
    resp = app_client.get("/search", params=params)
    assert resp.json()["context"]["returned"] == 0

    params["query"] = json.dumps(
        {"proj:epsg": {"eq": test_item["properties"]["proj:epsg"]}}
    )
    resp = app_client.get("/search", params=params)
    resp_json = resp.json()
    assert resp_json["context"]["returned"] == 1
    assert (
        resp_json["features"][0]["properties"]["proj:epsg"]
        == test_item["properties"]["proj:epsg"]
    )


def test_get_missing_item_collection(app_client):
    """Test reading a collection which does not exist"""
    resp = app_client.get("/collections/invalid-collection/items")
    assert resp.status_code == 200


def test_pagination_item_collection(app_client, load_test_data):
    """Test item collection pagination links (paging extension)"""
    test_item = load_test_data("test_item.json")
    ids = []

    # Ingest 5 items
    for idx in range(5):
        uid = str(uuid.uuid4())
        test_item["id"] = uid
        resp = app_client.post(
            f"/collections/{test_item['collection']}/items", json=test_item
        )
        assert resp.status_code == 200
        ids.append(uid)

    # Paginate through all 5 items with a limit of 1 (expecting 5 requests)
    page = app_client.get(
        f"/collections/{test_item['collection']}/items", params={"limit": 1}
    )
    idx = 0
    item_ids = []
    while True:
        idx += 1
        page_data = page.json()
        item_ids.append(page_data["features"][0]["id"])
        next_link = list(filter(lambda l: l["rel"] == "next", page_data["links"]))
        if not next_link:
            break
        query_params = parse_qs(urlparse(next_link[0]["href"]).query)
        page = app_client.get(
            f"/collections/{test_item['collection']}/items", params=query_params
        )

    # Our limit is 1 so we expect len(ids) number of requests before we run out of pages
    assert idx == len(ids)

    # Confirm we have paginated through all items
    assert not set(item_ids) - set(ids)


def test_pagination_post(app_client, load_test_data):
    """Test POST pagination (paging extension)"""
    test_item = load_test_data("test_item.json")
    ids = []

    # Ingest 5 items
    for idx in range(5):
        uid = str(uuid.uuid4())
        test_item["id"] = uid
        resp = app_client.post(
            f"/collections/{test_item['collection']}/items", json=test_item
        )
        assert resp.status_code == 200
        ids.append(uid)

    # Paginate through all 5 items with a limit of 1 (expecting 5 requests)
    request_body = {"ids": ids, "limit": 1}
    page = app_client.post("/search", json=request_body)
    idx = 0
    item_ids = []
    while True:
        idx += 1
        page_data = page.json()
        item_ids.append(page_data["features"][0]["id"])
        next_link = list(filter(lambda l: l["rel"] == "next", page_data["links"]))
        if not next_link:
            break
        # Merge request bodies
        request_body.update(next_link[0]["body"])
        page = app_client.post("/search", json=request_body)

    # Our limit is 1 so we expect len(ids) number of requests before we run out of pages
    assert idx == len(ids)

    # Confirm we have paginated through all items
    assert not set(item_ids) - set(ids)


def test_pagination_token_idempotent(app_client, load_test_data):
    """Test that pagination tokens are idempotent (paging extension)"""
    test_item = load_test_data("test_item.json")
    ids = []

    # Ingest 5 items
    for idx in range(5):
        uid = str(uuid.uuid4())
        test_item["id"] = uid
        resp = app_client.post(
            f"/collections/{test_item['collection']}/items", json=test_item
        )
        assert resp.status_code == 200
        ids.append(uid)

    page = app_client.get("/search", params={"ids": ",".join(ids), "limit": 3})
    page_data = page.json()
    next_link = list(filter(lambda l: l["rel"] == "next", page_data["links"]))

    # Confirm token is idempotent
    resp1 = app_client.get(
        "/search", params=parse_qs(urlparse(next_link[0]["href"]).query)
    )
    resp2 = app_client.get(
        "/search", params=parse_qs(urlparse(next_link[0]["href"]).query)
    )
    resp1_data = resp1.json()
    resp2_data = resp2.json()

    # Two different requests with the same pagination token should return the same items
    assert [item["id"] for item in resp1_data["features"]] == [
        item["id"] for item in resp2_data["features"]
    ]


def test_field_extension_get(app_client, load_test_data):
    """Test GET search with included fields (fields extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    params = {"fields": "+properties.proj:epsg,+properties.gsd"}
    resp = app_client.get("/search", params=params)
    feat_properties = resp.json()["features"][0]["properties"]
    assert not set(feat_properties) - {"proj:epsg", "gsd", "datetime"}


def test_field_extension_post(app_client, load_test_data):
    """Test POST search with included and excluded fields (fields extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    body = {
        "fields": {
            "exclude": ["assets.B1"],
            "include": ["properties.eo:cloud_cover", "properties.orientation"],
        }
    }

    resp = app_client.post("/search", json=body)
    resp_json = resp.json()
    assert "B1" not in resp_json["features"][0]["assets"].keys()
    assert not set(resp_json["features"][0]["properties"]) - {
        "orientation",
        "eo:cloud_cover",
        "datetime",
    }


def test_field_extension_exclude_and_include(app_client, load_test_data):
    """Test POST search including/excluding same field (fields extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    body = {
        "fields": {
            "exclude": ["properties.eo:cloud_cover"],
            "include": ["properties.eo:cloud_cover"],
        }
    }

    resp = app_client.post("/search", json=body)
    resp_json = resp.json()
    assert "eo:cloud_cover" not in resp_json["features"][0]["properties"]


def test_field_extension_exclude_default_includes(app_client, load_test_data):
    """Test POST search excluding a forbidden field (fields extension)"""
    test_item = load_test_data("test_item.json")
    resp = app_client.post(
        f"/collections/{test_item['collection']}/items", json=test_item
    )
    assert resp.status_code == 200

    body = {"fields": {"exclude": ["geometry"]}}

    resp = app_client.post("/search", json=body)
    resp_json = resp.json()
    assert "geometry" not in resp_json["features"][0]


def test_search_intersects_and_bbox(app_client):
    """Test POST search intersects and bbox are mutually exclusive (core)"""
    bbox = [-118, 34, -117, 35]
    geoj = Polygon.from_bounds(*bbox).__geo_interface__
    params = {"bbox": bbox, "intersects": geoj}
    resp = app_client.post("/search", json=params)
    assert resp.status_code == 422


def test_get_missing_item(app_client, load_test_data):
    """Test read item which does not exist (transactions extension)"""
    test_coll = load_test_data("test_collection.json")
    resp = app_client.get(f"/collections/{test_coll['id']}/items/invalid-item")
    assert resp.status_code == 404


def test_search_invalid_query_field(app_client):
    body = {"query": {"gsd": {"lt": 100}, "invalid-field": {"eq": 50}}}
    resp = app_client.post("/search", json=body)
    assert resp.status_code == 422



"""transactions extension client"""

import logging
from dataclasses import dataclass
from typing import Type, Union

from stac_api import errors
from stac_api.clients.base import BaseTransactionsClient
from stac_api.clients.postgres.base import PostgresClient
from stac_api.models import database, schemas

logger = logging.getLogger(__name__)


@dataclass
class TransactionsClient(PostgresClient, BaseTransactionsClient):
    """Transactions extension specific CRUD operations"""

    table: Type[database.Collection] = database.Collection
    item_table: Type[database.Item] = database.Item

    @property
    def collection_table(self):
        """alias for `self.table` # TODO: Figure out a better way to do this"""
        return self.table

    def _create(
        self,
        model: Union[schemas.Collection, schemas.Item],
        table: Union[Type[database.Collection], Type[database.Item]],
    ) -> Union[database.Collection, database.Item]:
        """Create a single record"""
        try:
            self.lookup_id(model.id, table=table)
            error_message = f"Row {model.id} already exists"
            logger.error(error_message, exc_info=True)
            raise errors.ConflictError(error_message)
        except errors.NotFoundError:
            row_data = table.from_schema(model)
            self.writer_session.add(row_data)
            self.commit()
            return row_data

    def _update(
        self,
        model: Union[schemas.Collection, schemas.Item],
        table: Union[Type[database.Collection], Type[database.Item]],
    ) -> Union[database.Collection, database.Item]:
        """Create a single record if it does not exist or update an existing record"""
        try:
            query = self.lookup_id(model.id, table=table)
            update_data = table.get_database_model(model)
            # SQLAlchemy orm updates don't seem to like geoalchemy types
            update_data.pop("geometry", None)
            query.update(update_data)
            self.commit()
            return table.from_schema(model)
        except errors.NotFoundError:
            row_data = table.from_schema(model)
            self.writer_session.add(row_data)
            self.commit()
            return row_data

    def _delete(
        self, item_id: str, table: Union[Type[database.Collection], Type[database.Item]]
    ) -> Union[database.Collection, database.Item]:
        """Delete a single record"""
        query = self.lookup_id(item_id, table=table)
        row_data = query.first()
        query.delete()
        self.commit()
        return row_data

    def create_item(self, model: schemas.Item, **kwargs) -> schemas.Item:
        """Create an item"""
        obj = self._create(model, table=self.item_table)
        obj.base_url = str(kwargs["request"].base_url)
        return schemas.Item.from_orm(obj)

    def update_item(self, model: schemas.Item, **kwargs) -> schemas.Item:
        """Update an item"""
        obj = self._update(model, table=self.item_table)
        obj.base_url = str(kwargs["request"].base_url)
        return schemas.Item.from_orm(obj)

    def delete_item(self, id: str, **kwargs) -> schemas.Item:
        """Delete an item"""
        obj = self._delete(id, table=self.item_table)
        obj.base_url = str(kwargs["request"].base_url)
        return schemas.Item.from_orm(obj)

    def create_collection(
        self, model: schemas.Collection, **kwargs
    ) -> schemas.Collection:
        """Create a collection"""
        obj = self._create(model, table=self.collection_table)
        obj.base_url = str(kwargs["request"].base_url)
        return schemas.Collection.from_orm(obj)

    def update_collection(
        self, model: schemas.Collection, **kwargs
    ) -> schemas.Collection:
        """Update a collection"""
        obj = self._update(model, table=self.collection_table)
        obj.base_url = str(kwargs["request"].base_url)
        return schemas.Collection.from_orm(obj)

    def delete_collection(self, id: str, **kwargs) -> schemas.Collection:
        """Delete a collection"""
        obj = self._delete(id, table=self.collection_table)
        obj.base_url = str(kwargs["request"].base_url)
        return schemas.Collection.from_orm(obj)



"""link helpers."""

from dataclasses import dataclass
from typing import Dict, List
from urllib.parse import urljoin

from stac_api import config
from stac_api.models.ogc import OGCTileLink
from stac_pydantic.shared import Link, MimeTypes, Relations

# These can be inferred from the item/collection so they aren't included in the database
# Instead they are dynamically generated when querying the database using the classes defined below
INFERRED_LINK_RELS = ["self", "item", "parent", "collection", "root"]


def filter_links(links: List[Dict]) -> List[Dict]:
    """Remove inferred links"""
    return [link for link in links if link["rel"] not in INFERRED_LINK_RELS]


@dataclass
class BaseLinks:
    """Create inferred links common to collections and items"""

    collection_id: str
    base_url: str

    def root(self) -> Link:
        """Return the catalog root"""
        return Link(
            rel=Relations.root, type=MimeTypes.json, href=urljoin(self.base_url, "/")
        )


@dataclass
class CollectionLinks(BaseLinks):
    """Create inferred links specific to collections"""

    def self(self) -> Link:
        """Create the `self` link."""
        return Link(
            rel=Relations.self,
            type=MimeTypes.json,
            href=urljoin(self.base_url, f"/collections/{self.collection_id}"),
        )

    def parent(self) -> Link:
        """Create the `parent` link."""
        return Link(
            rel=Relations.parent, type=MimeTypes.json, href=urljoin(self.base_url, "/")
        )

    def item(self) -> Link:
        """Create the `item` link."""
        return Link(
            rel=Relations.item,
            type=MimeTypes.geojson,
            href=urljoin(self.base_url, f"/collections/{self.collection_id}/items"),
        )

    def create_links(self) -> List[Link]:
        """Convenience method to return all inferred links"""
        return [self.self(), self.parent(), self.item(), self.root()]


@dataclass
class ItemLinks(BaseLinks):
    """Create inferred links specific to items"""

    item_id: str

    def self(self) -> Link:
        """Create the `self` link."""
        return Link(
            rel=Relations.self,
            type=MimeTypes.geojson,
            href=urljoin(
                self.base_url, f"/collections/{self.collection_id}/items/{self.item_id}"
            ),
        )

    def parent(self) -> Link:
        """Create the `parent` link."""
        return Link(
            rel=Relations.parent,
            type=MimeTypes.json,
            href=urljoin(self.base_url, f"/collections/{self.collection_id}"),
        )

    def collection(self) -> Link:
        """Create the `collection` link."""
        return Link(
            rel=Relations.collection,
            type=MimeTypes.json,
            href=urljoin(self.base_url, f"/collections/{self.collection_id}"),
        )

    def tiles(self) -> Link:
        """Create the `tiles` link"""
        return Link(
            rel=Relations.alternate,
            type=MimeTypes.json,
            title="tiles",
            href=urljoin(
                self.base_url,
                f"/collections/{self.collection_id}/items/{self.item_id}/tiles",
            ),
        )

    def create_links(self) -> List[Link]:
        """Convenience method to return all inferred links"""
        links = [
            self.self(),
            self.parent(),
            self.collection(),
            self.root(),
        ]
        if config.settings.add_on_is_enabled(config.AddOns.tiles):
            links.append(self.tiles())
        return links


@dataclass
class TileLinks:
    """Create inferred links specific to OGC Tiles API"""

    base_url: str
    collection_id: str
    item_id: str

    def __post_init__(self):
        """post init"""
        self.item_uri = urljoin(
            self.base_url, f"/collections/{self.collection_id}/items/{self.item_id}"
        )

    def tiles(self) -> OGCTileLink:
        """Create tiles link"""
        return OGCTileLink(
            href=urljoin(
                self.base_url,
                f"/titiler/tiles/{{z}}/{{x}}/{{y}}.png?url={self.item_uri}",
            ),
            rel=Relations.item,
            title="tiles",
            type=MimeTypes.png,
            templated=True,
        )

    def viewer(self) -> OGCTileLink:
        """Create viewer link"""
        return OGCTileLink(
            href=urljoin(self.base_url, f"/titiler/viewer?url={self.item_uri}"),
            rel=Relations.alternate,
            type=MimeTypes.html,
            title="viewer",
        )

    def tilejson(self) -> OGCTileLink:
        """Create tilejson link"""
        return OGCTileLink(
            href=urljoin(self.base_url, f"/titiler/tilejson.json?url={self.item_uri}"),
            rel=Relations.alternate,
            type=MimeTypes.json,
            title="tilejson",
        )

    def wmts(self) -> OGCTileLink:
        """Create wmts capabilities link"""
        return OGCTileLink(
            href=urljoin(
                self.base_url, f"/titiler/WMTSCapabilities.xml?url={self.item_uri}"
            ),
            rel=Relations.alternate,
            type=MimeTypes.xml,
            title="WMTS Capabilities",
        )

    def create_links(self) -> List[OGCTileLink]:
        """Convenience method to return all inferred links"""
        return [self.tiles(), self.tilejson(), self.wmts(), self.viewer()]



"""Item crud client."""
import json
import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional, Type, Union
from urllib.parse import urlencode, urljoin

import sqlalchemy as sa
from fastapi import Depends
from sqlalchemy import func

import geoalchemy2 as ga
from sqlakeyset import get_page
from stac_api import config, errors
from stac_api.clients.base import BaseCoreClient
from stac_api.clients.postgres.base import PostgresClient
from stac_api.clients.postgres.tokens import (
    PaginationTokenClient,
    pagination_token_client_factory,
)
from stac_api.config import ApiExtensions
from stac_api.errors import DatabaseError
from stac_api.models import database, schemas
from stac_api.models.links import CollectionLinks
from stac_pydantic import ItemCollection
from stac_pydantic.api import ConformanceClasses, LandingPage
from stac_pydantic.api.extensions.paging import PaginationLink
from stac_pydantic.shared import Link, MimeTypes, Relations

logger = logging.getLogger(__name__)

NumType = Union[float, int]


@dataclass
class CoreCrudClient(PostgresClient, BaseCoreClient):
    """Client for core endpoints defined by stac"""

    pagination_client: Optional[PaginationTokenClient] = None
    table: Type[database.Item] = database.Item
    collection_table: Type[database.Collection] = database.Collection

    def landing_page(self, **kwargs) -> LandingPage:
        """landing page"""
        landing_page = LandingPage(
            title="Arturo STAC API",
            description="Arturo raster datastore",
            links=[
                Link(
                    rel=Relations.self,
                    type=MimeTypes.json,
                    href=str(kwargs["request"].base_url),
                ),
                Link(
                    rel=Relations.docs,
                    type=MimeTypes.html,
                    title="OpenAPI docs",
                    href=urljoin(str(kwargs["request"].base_url), "/docs"),
                ),
                Link(
                    rel=Relations.conformance,
                    type=MimeTypes.json,
                    title="STAC/WFS3 conformance classes implemented by this server",
                    href=urljoin(str(kwargs["request"].base_url), "/conformance"),
                ),
                Link(
                    rel=Relations.search,
                    type=MimeTypes.geojson,
                    title="STAC search",
                    href=urljoin(str(kwargs["request"].base_url), "/search"),
                ),
            ],
        )
        collections = self.all_collections(request=kwargs["request"])
        for coll in collections:
            coll_link = CollectionLinks(
                collection_id=coll.id, base_url=str(kwargs["request"].base_url)
            ).self()
            coll_link.rel = Relations.child
            coll_link.title = coll.title
            landing_page.links.append(coll_link)
        return landing_page

    def conformance(self, **kwargs) -> ConformanceClasses:
        """conformance classes"""
        return ConformanceClasses(
            conformsTo=[
                "https://stacspec.org/STAC-api.html",
                "http://docs.opengeospatial.org/is/17-069r3/17-069r3.html#ats_geojson",
            ]
        )

    def all_collections(self, **kwargs) -> List[schemas.Collection]:
        """Read all collections from the database"""
        try:
            collections = self.reader_session.query(self.collection_table).all()
        except Exception as e:
            logger.error(e, exc_info=True)
            raise errors.DatabaseError(
                "Unhandled database error when getting item collection"
            )

        response = []
        for collection in collections:
            collection.base_url = str(kwargs["request"].base_url)
            response.append(schemas.Collection.from_orm(collection))
        return response

    def get_collection(self, id: str, **kwargs) -> schemas.Collection:
        """Get collection by id"""
        collection = self.lookup_id(id, table=self.collection_table).first()
        collection.base_url = str(kwargs["request"].base_url)
        return schemas.Collection.from_orm(collection)

    def item_collection(
        self, id: str, limit: int = 10, token: str = None, **kwargs
    ) -> ItemCollection:
        """Read an item collection from the database"""
        try:
            collection_children = (
                self.reader_session.query(self.table)
                .join(self.collection_table)
                .filter(self.collection_table.id == id)
                .order_by(self.table.datetime.desc(), self.table.id)
            )
            count = None
            if config.settings.api_extension_is_enabled(config.ApiExtensions.context):
                count_query = collection_children.statement.with_only_columns(
                    [func.count()]
                ).order_by(None)
                count = collection_children.session.execute(count_query).scalar()
            token = self.pagination_client.get(token) if token else token
            page = get_page(collection_children, per_page=limit, page=(token or False))
            # Create dynamic attributes for each page
            page.next = (
                self.pagination_client.insert(keyset=page.paging.bookmark_next)
                if page.paging.has_next
                else None
            )
            page.previous = (
                self.pagination_client.insert(keyset=page.paging.bookmark_previous)
                if page.paging.has_previous
                else None
            )
        except errors.NotFoundError:
            raise
        except Exception as e:
            logger.error(e, exc_info=True)
            raise errors.DatabaseError(
                "Unhandled database error when getting collection children"
            )

        links = []
        if page.next:
            links.append(
                PaginationLink(
                    rel=Relations.next,
                    type="application/geo+json",
                    href=f"{kwargs['request'].base_url}collections/{id}/items?token={page.next}&limit={limit}",
                    method="GET",
                )
            )
        if page.previous:
            links.append(
                PaginationLink(
                    rel=Relations.previous,
                    type="application/geo+json",
                    href=f"{kwargs['request'].base_url}collections/{id}/items?token={page.previous}&limit={limit}",
                    method="GET",
                )
            )

        response_features = []
        for item in page:
            item.base_url = str(kwargs["request"].base_url)
            response_features.append(schemas.Item.from_orm(item))

        context_obj = None
        if config.settings.api_extension_is_enabled(ApiExtensions.context):
            context_obj = {"returned": len(page), "limit": limit, "matched": count}

        return ItemCollection(
            type="FeatureCollection",
            context=context_obj,
            features=response_features,
            links=links,
        )

    def get_item(self, id: str, **kwargs) -> schemas.Item:
        """Get item by id"""
        obj = self.lookup_id(id).first()
        obj.base_url = str(kwargs["request"].base_url)
        return schemas.Item.from_orm(obj)

    def get_search(
        self,
        collections: Optional[List[str]] = None,
        ids: Optional[List[str]] = None,
        bbox: Optional[List[NumType]] = None,
        datetime: Optional[Union[str, datetime]] = None,
        limit: Optional[int] = 10,
        query: Optional[str] = None,
        token: Optional[str] = None,
        fields: Optional[List[str]] = None,
        sortby: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """GET search catalog"""
        # Parse request parameters
        base_args = {
            "collections": collections,
            "ids": ids,
            "bbox": bbox,
            "limit": limit,
            "token": token,
            "query": json.loads(query) if query else query,
        }
        if datetime:
            base_args["datetime"] = datetime
        if sortby:
            # https://github.com/radiantearth/stac-spec/tree/master/api-spec/extensions/sort#http-get-or-post-form
            sort_param = []
            for sort in sortby:
                sort_param.append(
                    {
                        "field": sort[1:],
                        "direction": "asc" if sort[0] == "+" else "desc",
                    }
                )
            base_args["sortby"] = sort_param

        if fields:
            includes = set()
            excludes = set()
            for field in fields:
                if field[0] == "-":
                    excludes.add(field[1:])
                elif field[0] == "+":
                    includes.add(field[1:])
                else:
                    includes.add(field)
            base_args["fields"] = {"include": includes, "exclude": excludes}

        # Do the request
        search_request = schemas.STACSearch(**base_args)
        resp = self.post_search(search_request, request=kwargs["request"])

        # Pagination
        page_links = []
        for link in resp["links"]:
            if link.rel == Relations.next or link.rel == Relations.previous:
                query_params = dict(kwargs["request"].query_params)
                if link.body and link.merge:
                    query_params.update(link.body)
                link.method = "GET"
                link.href = f"{link.href}?{urlencode(query_params)}"
                link.body = None
                link.merge = False
                page_links.append(link)
            else:
                page_links.append(link)
        resp["links"] = page_links
        return resp

    def post_search(
        self, search_request: schemas.STACSearch, **kwargs
    ) -> Dict[str, Any]:
        """POST search catalog"""
        token = (
            self.pagination_client.get(search_request.token)
            if search_request.token
            else False
        )
        query = self.reader_session.query(self.table)

        # Filter by collection
        count = None
        if search_request.collections:
            query = query.join(self.collection_table).filter(
                sa.or_(
                    *[
                        self.collection_table.id == col_id
                        for col_id in search_request.collections
                    ]
                )
            )

        # Sort
        if search_request.sortby:
            sort_fields = [
                getattr(self.table.get_field(sort.field), sort.direction.value)()
                for sort in search_request.sortby
            ]
            sort_fields.append(self.table.id)
            query = query.order_by(*sort_fields)
        else:
            # Default sort is date
            query = query.order_by(self.table.datetime.desc(), self.table.id)

        # Ignore other parameters if ID is present
        if search_request.ids:
            id_filter = sa.or_(*[self.table.id == i for i in search_request.ids])
            try:
                items = query.filter(id_filter).order_by(self.table.id)
                page = get_page(items, per_page=search_request.limit, page=token)
                if config.settings.api_extension_is_enabled(ApiExtensions.context):
                    count = len(search_request.ids)
                page.next = (
                    self.pagination_client.insert(keyset=page.paging.bookmark_next)
                    if page.paging.has_next
                    else None
                )
                page.previous = (
                    self.pagination_client.insert(keyset=page.paging.bookmark_previous)
                    if page.paging.has_previous
                    else None
                )
            except Exception as e:
                logger.error(e, exc_info=True)
                raise DatabaseError(
                    "Unhandled database error when searching for items by id"
                )
        else:
            # Spatial query
            poly = search_request.polygon()
            if poly:
                filter_geom = ga.shape.from_shape(poly, srid=4326)
                query = query.filter(
                    ga.func.ST_Intersects(self.table.geometry, filter_geom)
                )

            # Temporal query
            if search_request.datetime:
                # Two tailed query (between)
                if ".." not in search_request.datetime:
                    query = query.filter(
                        self.table.datetime.between(*search_request.datetime)
                    )
                # All items after the start date
                if search_request.datetime[0] != "..":
                    query = query.filter(
                        self.table.datetime >= search_request.datetime[0]
                    )
                # All items before the end date
                if search_request.datetime[1] != "..":
                    query = query.filter(
                        self.table.datetime <= search_request.datetime[1]
                    )

            # Query fields
            if search_request.query:
                for (field_name, expr) in search_request.query.items():
                    field = self.table.get_field(field_name)
                    for (op, value) in expr.items():
                        query = query.filter(op.operator(field, value))

            try:
                if config.settings.api_extension_is_enabled(ApiExtensions.context):
                    count_query = query.statement.with_only_columns(
                        [func.count()]
                    ).order_by(None)
                    count = query.session.execute(count_query).scalar()
                page = get_page(query, per_page=search_request.limit, page=token)
                # Create dynamic attributes for each page
                page.next = (
                    self.pagination_client.insert(keyset=page.paging.bookmark_next)
                    if page.paging.has_next
                    else None
                )
                page.previous = (
                    self.pagination_client.insert(keyset=page.paging.bookmark_previous)
                    if page.paging.has_previous
                    else None
                )
            except Exception as e:
                logger.error(e, exc_info=True)
                raise DatabaseError(
                    "Unhandled database error during spatial/temporal query"
                )
        links = []
        if page.next:
            links.append(
                PaginationLink(
                    rel=Relations.next,
                    type="application/geo+json",
                    href=f"{kwargs['request'].base_url}search",
                    method="POST",
                    body={"token": page.next},
                    merge=True,
                )
            )
        if page.previous:
            links.append(
                PaginationLink(
                    rel=Relations.previous,
                    type="application/geo+json",
                    href=f"{kwargs['request'].base_url}search",
                    method="POST",
                    body={"token": page.previous},
                    merge=True,
                )
            )

        response_features = []
        filter_kwargs = {}
        if config.settings.api_extension_is_enabled(ApiExtensions.fields):
            filter_kwargs = search_request.field.filter_fields

        xvals = []
        yvals = []
        for item in page:
            item.base_url = str(kwargs["request"].base_url)
            item_model = schemas.Item.from_orm(item)
            xvals += [item_model.bbox[0], item_model.bbox[2]]
            yvals += [item_model.bbox[1], item_model.bbox[3]]
            response_features.append(item_model.to_dict(**filter_kwargs))

        try:
            bbox = (min(xvals), min(yvals), max(xvals), max(yvals))
        except ValueError:
            bbox = None

        context_obj = None
        if config.settings.api_extension_is_enabled(ApiExtensions.context):
            context_obj = {
                "returned": len(page),
                "limit": search_request.limit,
                "matched": count,
            }

        return {
            "type": "FeatureCollection",
            "context": context_obj,
            "features": response_features,
            "links": links,
            "bbox": bbox,
        }


def core_crud_client_factory(
    pagination_client: PaginationTokenClient = Depends(pagination_token_client_factory),
) -> CoreCrudClient:
    """FastAPI dependency."""
    return CoreCrudClient(table=database.Item, pagination_client=pagination_client,)



"""ogc tiles client"""

from dataclasses import dataclass
from typing import Union

from starlette.responses import RedirectResponse

from stac_api.clients.postgres.core import CoreCrudClient
from stac_api.models.links import TileLinks
from stac_api.models.ogc import TileSetResource
from stac_pydantic.collection import SpatialExtent
from stac_pydantic.shared import MimeTypes


# TODO: Decouple from postgres by inherting from base class (stac_api.clients.base)
@dataclass
class TilesClient(CoreCrudClient):
    """OGC Tiles specific operations"""

    def get_item_tiles(
        self, id: str, **kwargs
    ) -> Union[RedirectResponse, TileSetResource]:
        """Get OGC TileSet resource for a stac item"""
        item = self.get_item(id, **kwargs)
        resource = TileSetResource(
            extent=SpatialExtent(bbox=[list(item.bbox)]),
            title=f"Tiled layer of {item.collection}/{item.id}",
            links=TileLinks(
                item_id=item.id,
                collection_id=item.collection,
                base_url=str(kwargs["request"].base_url),
            ).create_links(),
        )

        if "text/html" in kwargs["request"].headers["accept"]:
            viewer_url = [
                link.href for link in resource.links if link.type == MimeTypes.html
            ][0]
            return RedirectResponse(viewer_url)

        return resource



"""API pydantic models"""

import operator
from dataclasses import dataclass
from datetime import datetime
from enum import auto
from types import DynamicClassAttribute
from typing import Any, Callable, Dict, List, Optional, Set, Union

import sqlalchemy as sa
from shapely.geometry import Polygon as ShapelyPolygon
from shapely.geometry import shape

from geojson_pydantic.geometries import Polygon
from pydantic import Field, ValidationError, root_validator
from pydantic.error_wrappers import ErrorWrapper
from stac_api import config
from stac_api.models.decompose import CollectionGetter, ItemGetter
from stac_pydantic import Collection as CollectionBase
from stac_pydantic import Item as ItemBase
from stac_pydantic.api import Search
from stac_pydantic.api.extensions.fields import FieldsExtension as FieldsBase
from stac_pydantic.api.search import DATETIME_RFC339
from stac_pydantic.shared import Link
from stac_pydantic.utils import AutoValueEnum

# Be careful: https://github.com/samuelcolvin/pydantic/issues/1423#issuecomment-642797287
NumType = Union[float, int]


class Operator(str, AutoValueEnum):
    """
    Define our own operators because all operators defined in stac-pydantic are not currently supported.
    """

    eq = auto()
    ne = auto()
    lt = auto()
    le = auto()
    gt = auto()
    ge = auto()
    # TODO: These are defined in the spec but aren't currently implemented by the api
    # startsWith = auto()
    # endsWith = auto()
    # contains = auto()
    # in = auto()

    @DynamicClassAttribute
    def operator(self) -> Callable[[Any, Any], bool]:
        """Return python operator"""
        return getattr(operator, self._value_)


class Queryables(str, AutoValueEnum):
    """
    Define an enum of queryable fields and their data type.  Queryable fields are explicitly defined for two reasons:
        1. So the caller knows which fields they can query by
        2. Because JSONB queries with sqlalchemy ORM require casting the type of the field at runtime
            (see ``QueryableTypes``)

    # TODO: Let the user define these in a config file
    """

    orientation = auto()
    gsd = auto()
    epsg = "proj:epsg"
    height = auto()
    width = auto()
    minzoom = "cog:minzoom"
    maxzoom = "cog:maxzoom"
    dtype = "cog:dtype"


@dataclass
class QueryableTypes:
    """
    Define an enum of the field type of each queryable field

    # TODO: Let the user define these in a config file
    # TODO: There is a much better way of defining this field <> type mapping than two enums with same keys
    """

    orientation = sa.String
    gsd = sa.Float
    epsg = sa.Integer
    height = sa.Integer
    width = sa.Integer
    minzoom = sa.Integer
    maxzoom = sa.Integer
    dtype = sa.String


class FieldsExtension(FieldsBase):
    """Fields extension"""

    include: Optional[Set[str]] = set()
    exclude: Optional[Set[str]] = set()

    def _get_field_dict(self, fields: Set[str]) -> Dict:
        """
        Internal method to reate a dictionary for advanced include or exclude of pydantic fields on model export

        Ref: https://pydantic-docs.helpmanual.io/usage/exporting_models/#advanced-include-and-exclude
        """
        field_dict = {}
        for field in fields:
            if "." in field:
                parent, key = field.split(".")
                if parent not in field_dict:
                    field_dict[parent] = {key}
                else:
                    field_dict[parent].add(key)
            else:
                field_dict[field] = ...  # type:ignore
        return field_dict

    @property
    def filter_fields(self) -> Dict:
        """
        Create dictionary of fields to include/exclude on model export based on the included and excluded fields passed
        to the API

        Ref: https://pydantic-docs.helpmanual.io/usage/exporting_models/#advanced-include-and-exclude
        """
        # Include default set of fields
        include = config.settings.default_includes
        # If only include is specified, add fields to default set
        if self.include and not self.exclude:
            include = include.union(self.include)
        # If both include + exclude specified, find the difference between sets but don't remove any default fields
        # If we remove default fields we will get a validation error
        elif self.include and self.exclude:
            include = include.union(self.include) - (
                self.exclude - config.settings.default_includes
            )
        return {
            "include": self._get_field_dict(include),
            "exclude": self._get_field_dict(self.exclude),
        }


class Collection(CollectionBase):
    """Collection model"""

    links: Optional[List[Link]]

    class Config:
        """model config"""

        orm_mode = True
        use_enum_values = True
        getter_dict = CollectionGetter


class Item(ItemBase):
    """Item model"""

    geometry: Polygon
    links: Optional[List[Link]]

    class Config:
        """model config"""

        json_encoders = {datetime: lambda v: v.strftime(DATETIME_RFC339)}
        use_enum_values = True
        orm_mode = True
        getter_dict = ItemGetter


class STACSearch(Search):
    """Search model"""

    # Make collections optional, default to searching all collections if none are provided
    collections: Optional[List[str]] = None
    # Override default field extension to include default fields and pydantic includes/excludes factory
    field: FieldsExtension = Field(FieldsExtension(), alias="fields")
    # Override query extension with supported operators
    query: Optional[Dict[Queryables, Dict[Operator, Any]]]
    token: Optional[str] = None

    @root_validator(pre=True)
    def validate_query_fields(cls, values: Dict) -> Dict:
        """validate query fields"""
        if "query" in values and values["query"]:
            queryable_fields = Queryables.__members__.values()
            for field_name in values["query"]:
                if field_name not in queryable_fields:
                    raise ValidationError(
                        [
                            ErrorWrapper(
                                ValueError(f"Cannot search on field: {field_name}"),
                                "STACSearch",
                            )
                        ],
                        STACSearch,
                    )
        return values

    @root_validator
    def include_query_fields(cls, values: Dict) -> Dict:
        """
        Root validator to ensure query fields are included in the API response
        """
        if "query" in values and values["query"]:
            query_include = set(
                [
                    k.value
                    if k in config.settings.indexed_fields
                    else f"properties.{k.value}"
                    for k in values["query"]
                ]
            )
            if not values["field"].include:
                values["field"].include = query_include
            else:
                values["field"].include.union(query_include)
        return values

    def polygon(self) -> Optional[ShapelyPolygon]:
        """
        Convenience method to create a shapely polygon for the spatial query (either `intersects` or `bbox`)
        """
        if self.intersects:
            return shape(self.intersects)
        elif self.bbox:
            return ShapelyPolygon.from_bounds(*self.bbox)
        else:
            return None



def test_ping_no_param(app_client):
    """
    Test ping endpoint with a mocked client.
    Args:
        app_client (TestClient): mocked client fixture
    """
    res = app_client.get("/_mgmt/ping")
    assert res.status_code == 200
    assert res.json() == {"message": "PONG"}



"""Pagination token client."""

import os
from base64 import urlsafe_b64encode
from dataclasses import dataclass
from typing import Type

from stac_api.clients.postgres.base import PostgresClient
from stac_api.errors import DatabaseError
from stac_api.models import database


@dataclass
class PaginationTokenClient(PostgresClient):
    """Pagination token specific CRUD operations"""

    table: Type[database.PaginationToken] = database.PaginationToken

    def insert(self, keyset: str, tries: int = 0) -> str:  # type:ignore
        """Insert a keyset into the database"""
        # uid has collision chance of 1e-7 percent
        uid = urlsafe_b64encode(os.urandom(6)).decode()
        try:
            token = database.PaginationToken(id=uid, keyset=keyset)
            self.writer_session.add(token)
            self.commit()
            return uid
        except DatabaseError:
            # Try again if uid already exists in the database
            # TODO: Explicitely check for ConflictError (if insert fails for other reasons it should be raised)
            self.insert(keyset, tries=tries + 1)
            if tries > 5:
                raise

    def get(self, token_id: str) -> str:
        """Retrieve a keyset from the database"""
        row = self.lookup_id(token_id).first()
        return row.keyset


def pagination_token_client_factory() -> PaginationTokenClient:
    """FastAPI dependency"""
    return PaginationTokenClient()



"""router factory"""
from typing import List

import pkg_resources
from fastapi import APIRouter
from starlette.requests import Request
from starlette.responses import HTMLResponse
from starlette.templating import Jinja2Templates

from stac_api.api.models import (
    CollectionUri,
    EmptyRequest,
    ItemCollectionUri,
    ItemUri,
    SearchGetRequest,
    _create_request_model,
)
from stac_api.api.routes import create_endpoint_from_model, create_endpoint_with_depends
from stac_api.clients.base import BaseCoreClient, BaseTransactionsClient
from stac_api.config import ApiExtensions, ApiSettings
from stac_api.models import schemas
from stac_pydantic import ItemCollection
from stac_pydantic.api import ConformanceClasses, LandingPage


def create_tiles_router() -> APIRouter:
    """Create API router with OGC tiles endpoints"""
    from titiler.endpoints.stac import STACTiler

    template_dir = pkg_resources.resource_filename("titiler", "templates")
    templates = Jinja2Templates(directory=template_dir)

    titiler_router = STACTiler().router

    @titiler_router.get("/viewer", response_class=HTMLResponse)
    def stac_demo(request: Request):
        """STAC Viewer."""
        return templates.TemplateResponse(
            name="stac_index.html",
            context={
                "request": request,
                "tilejson": request.url_for("tilejson"),
                "metadata": request.url_for("info"),
            },
            media_type="text/html",
        )

    # TODO: add titiler exception handlers
    return titiler_router


def create_core_router(client: BaseCoreClient, settings: ApiSettings) -> APIRouter:
    """Create API router with item endpoints"""
    search_request_model = _create_request_model(schemas.STACSearch, settings)
    router = APIRouter()
    router.add_api_route(
        name="Landing Page",
        path="/",
        response_model=LandingPage,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["GET"],
        endpoint=create_endpoint_with_depends(client.landing_page, EmptyRequest),
    )
    router.add_api_route(
        name="Conformance Classes",
        path="/conformance",
        response_model=ConformanceClasses,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["GET"],
        endpoint=create_endpoint_with_depends(client.conformance, EmptyRequest),
    )
    router.add_api_route(
        name="Get Item",
        path="/collections/{collectionId}/items/{itemId}",
        response_model=schemas.Item,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["GET"],
        endpoint=create_endpoint_with_depends(client.get_item, ItemUri),
    )
    router.add_api_route(
        name="Search",
        path="/search",
        response_model=None
        if settings.api_extension_is_enabled(ApiExtensions.fields)
        else ItemCollection,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["POST"],
        endpoint=create_endpoint_from_model(client.post_search, search_request_model),
    ),
    router.add_api_route(
        name="Search",
        path="/search",
        response_model=None
        if settings.api_extension_is_enabled(ApiExtensions.fields)
        else ItemCollection,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["GET"],
        endpoint=create_endpoint_with_depends(client.get_search, SearchGetRequest),
    )
    router.add_api_route(
        name="Get Collections",
        path="/collections",
        response_model=List[schemas.Collection],
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["GET"],
        endpoint=create_endpoint_with_depends(client.all_collections, EmptyRequest),
    )
    router.add_api_route(
        name="Get Collection",
        path="/collections/{collectionId}",
        response_model=schemas.Collection,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["GET"],
        endpoint=create_endpoint_with_depends(client.get_collection, CollectionUri),
    )
    router.add_api_route(
        name="Get ItemCollection",
        path="/collections/{collectionId}/items",
        response_model=ItemCollection,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["GET"],
        endpoint=create_endpoint_with_depends(
            client.item_collection, ItemCollectionUri
        ),
    )
    return router


def create_transactions_router(
    client: BaseTransactionsClient, settings: ApiSettings
) -> APIRouter:
    """Create API router for transactions extension"""
    item_request_model = _create_request_model(schemas.Item, settings)
    collection_request_model = _create_request_model(schemas.Collection, settings)
    router = APIRouter()
    router.add_api_route(
        name="Create Item",
        path="/collections/{collectionId}/items",
        response_model=schemas.Item,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["POST"],
        endpoint=create_endpoint_from_model(client.create_item, item_request_model),
    )
    router.add_api_route(
        name="Update Item",
        path="/collections/{collectionId}/items",
        response_model=schemas.Item,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["PUT"],
        endpoint=create_endpoint_from_model(client.update_item, item_request_model),
    )
    router.add_api_route(
        name="Delete Item",
        path="/collections/{collectionId}/items/{itemId}",
        response_model=schemas.Item,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["DELETE"],
        endpoint=create_endpoint_with_depends(client.delete_item, ItemUri),
    )
    router.add_api_route(
        name="Create Collection",
        path="/collections",
        response_model=schemas.Collection,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["POST"],
        endpoint=create_endpoint_from_model(
            client.create_collection, collection_request_model
        ),
    )
    router.add_api_route(
        name="Update Collection",
        path="/collections",
        response_model=schemas.Collection,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["PUT"],
        endpoint=create_endpoint_from_model(
            client.update_collection, collection_request_model
        ),
    )
    router.add_api_route(
        name="Delete Collection",
        path="/collections/{collectionId}",
        response_model=schemas.Collection,
        response_model_exclude_unset=True,
        response_model_exclude_none=True,
        methods=["DELETE"],
        endpoint=create_endpoint_with_depends(client.delete_collection, CollectionUri),
    )
    return router



"""openapi"""
from fastapi import FastAPI
from fastapi.openapi.utils import get_openapi

from stac_api.config import ApiExtensions, ApiSettings


def config_openapi(app: FastAPI, settings: ApiSettings):
    """config openapi"""

    def custom_openapi():
        """config openapi"""
        if app.openapi_schema:
            return app.openapi_schema

        openapi_schema = get_openapi(
            title="Arturo STAC API", version="0.1", routes=app.routes
        )

        if settings.api_extension_is_enabled(ApiExtensions.fields):
            openapi_schema["paths"]["/search"]["get"]["responses"]["200"]["content"][
                "application/json"
            ]["schema"] = {"$ref": "#/components/schemas/ItemCollection"}
            openapi_schema["paths"]["/search"]["post"]["responses"]["200"]["content"][
                "application/json"
            ]["schema"] = {"$ref": "#/components/schemas/ItemCollection"}

        app.openapi_schema = openapi_schema
        return app.openapi_schema

    app.openapi = custom_openapi



"""SQLAlchemy ORM models."""

import json
from datetime import datetime
from typing import Optional

import sqlalchemy as sa
from shapely.geometry import shape
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base

import geoalchemy2 as ga
from stac_api import config
from stac_api.models import schemas
from stac_pydantic.shared import DATETIME_RFC339

BaseModel = declarative_base()


class GeojsonGeometry(ga.Geometry):
    """
    Custom geoalchemy type which returns GeoJSON
    """

    from_text = "ST_GeomFromGeoJSON"

    def result_processor(self, dialect: str, coltype):
        """Override default processer to return GeoJSON"""

        def process(value: Optional[bytes]):
            if value is not None:
                geom = ga.shape.to_shape(
                    ga.elements.WKBElement(
                        value, srid=self.srid, extended=self.extended
                    )
                )
                return json.loads(json.dumps(geom.__geo_interface__))

        return process


class Collection(BaseModel):  # type:ignore
    """
    Collection orm model.
    """

    __tablename__ = "collections"
    __table_args__ = {"schema": "data"}

    id = sa.Column(sa.VARCHAR(1024), nullable=False, primary_key=True)
    stac_version = sa.Column(sa.VARCHAR(300))
    stac_extensions = sa.Column(sa.ARRAY(sa.VARCHAR(300)), nullable=True)
    title = sa.Column(sa.VARCHAR(1024))
    description = sa.Column(sa.VARCHAR(1024), nullable=False)
    keywords = sa.Column(sa.VARCHAR(300))
    version = sa.Column(sa.VARCHAR(300))
    license = sa.Column(sa.VARCHAR(300), nullable=False)
    providers = sa.Column(JSONB)
    summaries = sa.Column(JSONB, nullable=True)
    extent = sa.Column(JSONB)
    links = sa.Column(JSONB)
    children = sa.orm.relationship("Item", lazy="dynamic")

    @classmethod
    def get_database_model(cls, schema: schemas.Collection) -> dict:
        """Decompose pydantic model to data model"""
        return schema.dict(exclude_none=True)

    @classmethod
    def from_schema(cls, schema: schemas.Collection) -> "Collection":
        """Create orm model from pydantic model"""
        return cls(**cls.get_database_model(schema))


class Item(BaseModel):  # type:ignore
    """
    Item orm model.
    """

    __tablename__ = "items"
    __table_args__ = {"schema": "data"}

    id = sa.Column(sa.VARCHAR(1024), nullable=False, primary_key=True)
    stac_version = sa.Column(sa.VARCHAR(300))
    stac_extensions = sa.Column(sa.ARRAY(sa.VARCHAR(300)), nullable=True)
    geometry = sa.Column(GeojsonGeometry("POLYGON", srid=4326, spatial_index=True))
    bbox = sa.Column(sa.ARRAY(sa.NUMERIC), nullable=False)
    properties = sa.Column(JSONB)
    assets = sa.Column(JSONB)
    collection_id = sa.Column(
        sa.VARCHAR(1024), sa.ForeignKey(Collection.id), nullable=False
    )
    parent_collection = sa.orm.relationship("Collection", back_populates="children")
    datetime = sa.Column(sa.TIMESTAMP, nullable=False)
    links = sa.Column(JSONB)

    @classmethod
    def get_database_model(cls, schema: schemas.Item) -> dict:
        """Decompose pydantic model to data model"""
        indexed_fields = {}
        for field in config.settings.indexed_fields:
            # Use getattr to accommodate extension namespaces
            field_value = getattr(schema.properties, field)
            if field == "datetime":
                field_value = datetime.strptime(field_value, DATETIME_RFC339)
            indexed_fields[field.split(":")[-1]] = field_value

        # Exclude indexed fields from the properties jsonb field
        properties = schema.properties.dict(exclude=set(config.settings.indexed_fields))
        now = datetime.utcnow().strftime(DATETIME_RFC339)
        if not properties["created"]:
            properties["created"] = now
        properties["updated"] = now

        return dict(
            collection_id=schema.collection,
            geometry=ga.shape.from_shape(shape(schema.geometry), 4326),
            properties=properties,
            **indexed_fields,
            **schema.dict(
                exclude_none=True,
                exclude=set(
                    config.settings.forbidden_fields
                    | {"geometry", "properties", "collection"}
                ),
            )
        )

    @classmethod
    def from_schema(cls, schema: schemas.Item) -> "Item":
        """Create orm model from pydantic model"""
        return cls(**cls.get_database_model(schema))

    @classmethod
    def get_field(cls, field_name):
        """Get a model field"""
        try:
            return getattr(cls, field_name)
        except AttributeError:
            # Use a JSONB field
            return cls.properties[(field_name)].cast(
                getattr(schemas.QueryableTypes, field_name.name)
            )


class PaginationToken(BaseModel):  # type:ignore
    """
    Pagination orm model.
    """

    __tablename__ = "tokens"
    __table_args__ = {"schema": "data"}

    id = sa.Column(sa.VARCHAR(100), nullable=False, primary_key=True)
    keyset = sa.Column(sa.VARCHAR(1000), nullable=False)



def test_create_and_delete_collection(app_client, load_test_data):
    """Test creation and deletion of a collection"""
    test_collection = load_test_data("test_collection.json")
    test_collection["id"] = "test"

    resp = app_client.post("/collections", json=test_collection)
    assert resp.status_code == 200

    resp = app_client.delete(f"/collections/{test_collection['id']}")
    assert resp.status_code == 200


def test_create_collection_conflict(app_client, load_test_data):
    """Test creation of a collection which already exists"""
    # This collection ID is created in the fixture, so this should be a conflict
    test_collection = load_test_data("test_collection.json")
    resp = app_client.post("/collections", json=test_collection)
    assert resp.status_code == 409


def test_delete_missing_collection(app_client):
    """Test deletion of a collection which does not exist"""
    resp = app_client.delete("/collections/missing-collection")
    assert resp.status_code == 404


def test_update_collection_already_exists(app_client, load_test_data):
    """Test updating a collection which already exists"""
    test_collection = load_test_data("test_collection.json")
    test_collection["keywords"].append("test")
    resp = app_client.put("/collections", json=test_collection)
    assert resp.status_code == 200

    resp = app_client.get(f"/collections/{test_collection['id']}")
    assert resp.status_code == 200
    resp_json = resp.json()
    assert "test" in resp_json["keywords"]


def test_update_new_collection(app_client, load_test_data):
    """Test updating a collection which does not exist (same as creation)"""
    test_collection = load_test_data("test_collection.json")
    test_collection["id"] = "new-test-collection"

    resp = app_client.put("/collections", json=test_collection)
    assert resp.status_code == 200

    resp = app_client.get(f"/collections/{test_collection['id']}")
    assert resp.status_code == 200
    resp_json = resp.json()
    assert resp_json["id"] == test_collection["id"]


def test_collection_not_found(app_client):
    """Test read a collection which does not exist"""
    resp = app_client.get("/collections/does-not-exist")
    assert resp.status_code == 404



"""ogc tiles api models"""

from typing import List, Optional

from pydantic import BaseModel
from stac_pydantic.collection import SpatialExtent
from stac_pydantic.shared import Link


class OGCTileLink(Link):
    """OGC Tiles API - Link"""

    templated: Optional[bool] = False


class TileSetResource(BaseModel):
    """OGC Tiles API - TileSetResource"""

    extent: SpatialExtent
    links: List[OGCTileLink]
    title: Optional[str]
    description: Optional[str]



"""Base clients."""
import abc
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from stac_api.models import schemas
from stac_pydantic import ItemCollection
from stac_pydantic.api import ConformanceClasses, LandingPage

NumType = Union[float, int]


@dataclass  # type:ignore
class BaseTransactionsClient(abc.ABC):
    """Base transactions client"""

    @abc.abstractmethod
    def create_item(self, model: schemas.Item, **kwargs) -> schemas.Item:
        """Create item"""
        ...

    @abc.abstractmethod
    def update_item(self, model: schemas.Item, **kwargs) -> schemas.Item:
        """Update item"""
        ...

    @abc.abstractmethod
    def delete_item(self, id: str, **kwargs) -> schemas.Item:
        """Delete item"""
        ...

    @abc.abstractmethod
    def create_collection(
        self, model: schemas.Collection, **kwargs
    ) -> schemas.Collection:
        """Create collection"""
        ...

    @abc.abstractmethod
    def update_collection(
        self, model: schemas.Collection, **kwargs
    ) -> schemas.Collection:
        """Update collection"""
        ...

    @abc.abstractmethod
    def delete_collection(self, id: str, **kwargs) -> schemas.Collection:
        """Delete collection"""
        ...


@dataclass  # type:ignore
class BaseCoreClient(abc.ABC):
    """Base client for core endpoints defined by stac"""

    @abc.abstractmethod
    def landing_page(self, **kwargs) -> LandingPage:
        """landing page"""
        ...

    @abc.abstractmethod
    def conformance(self, **kwargs) -> ConformanceClasses:
        """conformance classes"""
        ...

    @abc.abstractmethod
    def post_search(
        self, search_request: schemas.STACSearch, **kwargs
    ) -> Dict[str, Any]:
        """search for items"""
        ...

    @abc.abstractmethod
    def get_search(
        self,
        collections: Optional[List[str]] = None,
        ids: Optional[List[str]] = None,
        bbox: Optional[List[NumType]] = None,
        datetime: Optional[Union[str, datetime]] = None,
        limit: Optional[int] = 10,
        query: Optional[str] = None,
        token: Optional[str] = None,
        fields: Optional[List[str]] = None,
        sortby: Optional[str] = None,
        **kwargs
    ) -> Dict[str, Any]:
        """GET search catalog"""
        ...

    @abc.abstractmethod
    def get_item(self, id: str, **kwargs) -> schemas.Item:
        """get item by id"""
        ...

    @abc.abstractmethod
    def all_collections(self) -> List[schemas.Collection]:
        """get all collections"""
        ...

    @abc.abstractmethod
    def get_collection(self, id: str, **kwargs) -> schemas.Collection:
        """get collection by id"""
        ...

    @abc.abstractmethod
    def item_collection(
        self, id: str, limit: int = 10, token: str = None, **kwargs
    ) -> ItemCollection:
        """Get item collection"""
        ...



"""Error handling"""

import logging
from typing import Callable, Dict, Type

from fastapi import FastAPI
from starlette import status
from starlette.requests import Request
from starlette.responses import JSONResponse

logger = logging.getLogger(__name__)


class StacApiError(Exception):
    """Generic API error"""

    pass


class ConflictError(StacApiError):
    """Database conflict"""

    pass


class NotFoundError(StacApiError):
    """Resource not found"""

    pass


class ForeignKeyError(StacApiError):
    """Foreign key error (collection does not exist)"""

    pass


class DatabaseError(StacApiError):
    """Generic database errors"""

    pass


DEFAULT_STATUS_CODES = {
    NotFoundError: status.HTTP_404_NOT_FOUND,
    ConflictError: status.HTTP_409_CONFLICT,
    ForeignKeyError: status.HTTP_422_UNPROCESSABLE_ENTITY,
    DatabaseError: status.HTTP_424_FAILED_DEPENDENCY,
    Exception: status.HTTP_500_INTERNAL_SERVER_ERROR,
}


def exception_handler_factory(status_code: int) -> Callable:
    """
    Create a FastAPI exception handler from a status code.
    """

    def handler(request: Request, exc: Exception):
        logger.error(exc, exc_info=True)
        return JSONResponse(content={"detail": str(exc)}, status_code=status_code)

    return handler


def add_exception_handlers(
    app: FastAPI, status_codes: Dict[Type[Exception], int]
) -> None:
    """
    Add exception handlers to the FastAPI app.
    """
    for (exc, code) in status_codes.items():
        app.add_exception_handler(exc, exception_handler_factory(code))



"""fastapi app creation"""

from fastapi import APIRouter, FastAPI
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from starlette.requests import Request

from stac_api.api.models import ItemUri
from stac_api.api.routers import (
    create_core_router,
    create_tiles_router,
    create_transactions_router,
)
from stac_api.api.routes import create_endpoint_with_depends
from stac_api.clients.postgres.core import CoreCrudClient
from stac_api.clients.postgres.tokens import PaginationTokenClient
from stac_api.clients.postgres.transactions import TransactionsClient
from stac_api.clients.tiles.ogc import TilesClient
from stac_api.config import AddOns, ApiExtensions, ApiSettings, inject_settings
from stac_api.errors import DEFAULT_STATUS_CODES, add_exception_handlers
from stac_api.models.ogc import TileSetResource
from stac_api.openapi import config_openapi
from stac_api.utils.dependencies import READER, WRITER


def create_app(settings: ApiSettings) -> FastAPI:
    """Create a FastAPI app"""
    paging_client = PaginationTokenClient()
    core_client = CoreCrudClient(pagination_client=paging_client)

    app = FastAPI()
    inject_settings(settings)

    app.debug = settings.debug
    app.include_router(
        create_core_router(core_client, settings), tags=["Core Endpoints"]
    )
    add_exception_handlers(app, DEFAULT_STATUS_CODES)

    if settings.api_extension_is_enabled(ApiExtensions.transaction):
        transaction_client = TransactionsClient()
        app.include_router(
            create_transactions_router(transaction_client, settings),
            tags=["Transaction Extension"],
        )

    if settings.add_on_is_enabled(AddOns.tiles):
        tiles_client = TilesClient()
        app.add_api_route(
            name="Get OGC Tiles Resource",
            path="/collections/{collectionId}/items/{itemId}/tiles",
            response_model=TileSetResource,
            response_model_exclude_none=True,
            response_model_exclude_unset=True,
            methods=["GET"],
            endpoint=create_endpoint_with_depends(tiles_client.get_item_tiles, ItemUri),
            tags=["OGC Tiles"],
        )
        app.include_router(create_tiles_router(), prefix="/titiler", tags=["Titiler"])

    config_openapi(app, settings)

    @app.on_event("startup")
    async def on_startup():
        """Create database engines and sessions on startup"""
        app.state.ENGINE_READER = create_engine(
            settings.reader_connection_string, echo=settings.debug
        )
        app.state.ENGINE_WRITER = create_engine(
            settings.writer_connection_string, echo=settings.debug
        )
        app.state.DB_READER = sessionmaker(
            autocommit=False, autoflush=False, bind=app.state.ENGINE_READER
        )
        app.state.DB_WRITER = sessionmaker(
            autocommit=False, autoflush=False, bind=app.state.ENGINE_WRITER
        )

    @app.on_event("shutdown")
    async def on_shutdown():
        """Dispose of database engines and sessions on app shutdown"""
        app.state.ENGINE_READER.dispose()
        app.state.ENGINE_WRITER.dispose()

    @app.middleware("http")
    async def create_db_connection(request: Request, call_next):
        """Create a new database connection for each request"""
        if "titiler" in str(request.url):
            return await call_next(request)
        reader = request.app.state.DB_READER()
        writer = request.app.state.DB_WRITER()
        READER.set(reader)
        WRITER.set(writer)
        resp = await call_next(request)
        reader.close()
        writer.close()
        return resp

    mgmt_router = APIRouter()

    @mgmt_router.get("/_mgmt/ping")
    async def ping():
        """Liveliness/readiness probe"""
        return {"message": "PONG"}

    app.include_router(mgmt_router, tags=["Liveliness/Readiness"])

    return app



"""api module"""
from .app import create_app
from .routers import create_core_router, create_tiles_router, create_transactions_router
from .routes import create_endpoint_from_model, create_endpoint_with_depends



import uuid
from typing import Callable

import pytest

from stac_api.clients.postgres.core import CoreCrudClient
from stac_api.clients.postgres.transactions import TransactionsClient
from stac_api.errors import ConflictError, NotFoundError
from stac_api.models.schemas import Collection, Item
from tests.conftest import MockStarletteRequest


def test_create_collection(
    postgres_core: CoreCrudClient,
    postgres_transactions: TransactionsClient,
    load_test_data: Callable,
):
    data = Collection.parse_obj(load_test_data("test_collection.json"))
    resp = postgres_transactions.create_collection(data, request=MockStarletteRequest)
    assert data.dict(exclude={"links"}) == resp.dict(exclude={"links"})
    coll = postgres_core.get_collection(data.id, request=MockStarletteRequest)
    assert coll.id == data.id


def test_create_collection_already_exists(
    postgres_core: CoreCrudClient,
    postgres_transactions: TransactionsClient,
    load_test_data: Callable,
):
    data = Collection.parse_obj(load_test_data("test_collection.json"))
    postgres_transactions.create_collection(data, request=MockStarletteRequest)

    with pytest.raises(ConflictError):
        postgres_transactions.create_collection(data, request=MockStarletteRequest)


def test_update_collection(
    postgres_core: CoreCrudClient,
    postgres_transactions: TransactionsClient,
    load_test_data: Callable,
):
    data = Collection.parse_obj(load_test_data("test_collection.json"))
    postgres_transactions.create_collection(data, request=MockStarletteRequest)

    data.keywords.append("new keyword")
    postgres_transactions.update_collection(data, request=MockStarletteRequest)

    coll = postgres_core.get_collection(data.id, request=MockStarletteRequest)
    assert "new keyword" in coll.keywords


def test_delete_collection(
    postgres_core: CoreCrudClient,
    postgres_transactions: TransactionsClient,
    load_test_data: Callable,
):
    data = Collection.parse_obj(load_test_data("test_collection.json"))
    postgres_transactions.create_collection(data, request=MockStarletteRequest)

    deleted = postgres_transactions.delete_collection(
        data.id, request=MockStarletteRequest
    )

    with pytest.raises(NotFoundError):
        postgres_core.get_collection(deleted.id, request=MockStarletteRequest)


def test_get_collection(
    postgres_core: CoreCrudClient,
    postgres_transactions: TransactionsClient,
    load_test_data: Callable,
):
    data = Collection.parse_obj(load_test_data("test_collection.json"))
    postgres_transactions.create_collection(data, request=MockStarletteRequest)
    coll = postgres_core.get_collection(data.id, request=MockStarletteRequest)
    assert data.dict(exclude={"links"}) == coll.dict(exclude={"links"})
    assert coll.id == data.id


def test_get_collection_items(
    postgres_core: CoreCrudClient,
    postgres_transactions: TransactionsClient,
    load_test_data: Callable,
):
    coll = Collection.parse_obj(load_test_data("test_collection.json"))
    postgres_transactions.create_collection(coll, request=MockStarletteRequest)

    item = Item.parse_obj(load_test_data("test_item.json"))

    for _ in range(5):
        item.id = str(uuid.uuid4())
        postgres_transactions.create_item(item, request=MockStarletteRequest)

    fc = postgres_core.item_collection(coll.id, request=MockStarletteRequest)
    assert len(fc.features) == 5

    for item in fc.features:
        assert item.collection == coll.id


def test_create_item(
    postgres_core: CoreCrudClient,
    postgres_transactions: TransactionsClient,
    load_test_data: Callable,
):
    coll = Collection.parse_obj(load_test_data("test_collection.json"))
    postgres_transactions.create_collection(coll, request=MockStarletteRequest)
    item = Item.parse_obj(load_test_data("test_item.json"))
    postgres_transactions.create_item(item, request=MockStarletteRequest)
    resp = postgres_core.get_item(item.id, request=MockStarletteRequest)
    assert item.dict(
        exclude={"links": ..., "properties": {"created", "updated"}}
    ) == resp.dict(exclude={"links": ..., "properties": {"created", "updated"}})


def test_create_item_already_exists(
    postgres_core: CoreCrudClient,
    postgres_transactions: TransactionsClient,
    load_test_data: Callable,
):
    coll = Collection.parse_obj(load_test_data("test_collection.json"))
    postgres_transactions.create_collection(coll, request=MockStarletteRequest)

    item = Item.parse_obj(load_test_data("test_item.json"))
    postgres_transactions.create_item(item, request=MockStarletteRequest)

    with pytest.raises(ConflictError):
        postgres_transactions.create_item(item, request=MockStarletteRequest)


def test_update_item(
    postgres_core: CoreCrudClient,
    postgres_transactions: TransactionsClient,
    load_test_data: Callable,
):
    coll = Collection.parse_obj(load_test_data("test_collection.json"))
    postgres_transactions.create_collection(coll, request=MockStarletteRequest)

    item = Item.parse_obj(load_test_data("test_item.json"))
    postgres_transactions.create_item(item, request=MockStarletteRequest)

    item.properties.foo = "bar"
    postgres_transactions.update_item(item, request=MockStarletteRequest)

    updated_item = postgres_core.get_item(item.id, request=MockStarletteRequest)
    assert updated_item.properties.foo == "bar"


def test_delete_item(
    postgres_core: CoreCrudClient,
    postgres_transactions: TransactionsClient,
    load_test_data: Callable,
):
    coll = Collection.parse_obj(load_test_data("test_collection.json"))
    postgres_transactions.create_collection(coll, request=MockStarletteRequest)

    item = Item.parse_obj(load_test_data("test_item.json"))
    postgres_transactions.create_item(item, request=MockStarletteRequest)

    postgres_transactions.delete_item(item.id, request=MockStarletteRequest)

    with pytest.raises(NotFoundError):
        postgres_core.get_item(item.id, request=MockStarletteRequest)



"""route factories"""
from typing import Callable, Type

from fastapi import Depends
from starlette.requests import Request

from pydantic import BaseModel
from stac_api.api.models import APIRequest


# TODO: Only use one endpoint factory
def create_endpoint_from_model(
    func: Callable, request_model: Type[BaseModel]
) -> Callable:
    """
    Create a FastAPI endpoint where request model is a pydantic model.  This works best for validating request bodies
    (POST/PUT etc.)
    """

    def _endpoint(
        request: Request, request_data: request_model,  # type:ignore
    ):
        """endpoint"""
        resp = func(request_data, request=request)
        return resp

    return _endpoint


def create_endpoint_with_depends(
    func: Callable, request_model: Type[APIRequest],
) -> Callable:
    """
    Create a fastapi endpoint where request model is a dataclass.  This works best for validating query/patm params.
    """

    def _endpoint(
        request: Request, request_data: request_model = Depends(),  # type:ignore
    ):
        """endpoint"""
        resp = func(
            request=request, **request_data.kwargs()  # type:ignore
        )
        return resp

    return _endpoint



"""arturo-stac-api"""

from setuptools import find_packages, setup

with open("README.md") as f:
    desc = f.read()


install_requires = [
    "uvicorn>=0.11.7",
    "gunicorn",
    "fastapi>=0.60.0",
    "alembic",
    "psycopg2-binary",
    "shapely",
    "sqlalchemy",
    "geoalchemy2<0.8.0",
    "sqlakeyset",
    "stac-pydantic>=1.3.5",
    "pydantic[dotenv]",
    "cogeo-mosaic==3.0a10",
    "titiler==0.1a2",
    "rio-cogeo==2.0a5",
]

extra_reqs = {
    "dev": ["pytest", "pytest-cov", "pytest-asyncio", "pre-commit", "requests"],
}


setup(
    name="arturo-stac-api",
    description="Arturo's STAC compliant API implementation",
    long_description=desc,
    long_description_content_type="text/markdown",
    version="1.0.0",
    python_requires=">=3.8",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
    keywords="STAC FastAPI COG",
    author=u"Arturo Engineering",
    author_email="engineering@arturo.ai",
    url="https://github.com/arturo-ai/arturo-stac-api",
    license="MIT",
    packages=find_packages(exclude=["alembic", "tests", "scripts"]),
    zip_safe=False,
    install_requires=install_requires,
    tests_require=extra_reqs["dev"],
    extras_require=extra_reqs,
)
